"""Tests for the qtpng preprocessor"""

# Copyright (c) IPython Development Team.
# Distributed under the terms of the Modified BSD License.

import os

import pytest
from flaky import flaky

from nbconvert.exporters.qt_screenshot import QT_INSTALLED
from nbconvert.exporters.qtpng import QtPNGExporter

from .base import ExportersTestsBase


@pytest.mark.skipif(not QT_INSTALLED, reason="PyQtWebEngine not installed")
class TestQtPNGExporter(ExportersTestsBase):
    """Contains test functions for qtpng.py"""

    exporter_class = QtPNGExporter  # type:ignore

    @flaky
    def test_export(self):
        """
        Can a TemplateExporter export something?
        """
        if os.name == "nt":
            # currently not supported
            with pytest.raises(RuntimeError):
                (output, resources) = QtPNGExporter().from_filename(self._get_notebook())
        else:
            (output, resources) = QtPNGExporter().from_filename(self._get_notebook())
            assert len(output) > 0



from nbconvert.utils.version import check_version


def test_check_version():
    """Test the behaviour of check_version.

    This is mostly used to make sure the pandoc version is appropriate for the library.
    """

    assert check_version("1.19.2.4", "1.12.1")
    assert check_version("2.2.3.2", "1.12.1")
    assert check_version("1.19.2.4", "1.12.1", max_v="2.0")
    assert not check_version("2.2.3.2", "1.12.1", max_v="2.0")



"""Tests for SlidesExporter"""

# Copyright (c) IPython Development Team.
# Distributed under the terms of the Modified BSD License.

from nbformat import v4 as nbformat

from nbconvert.exporters.slides import SlidesExporter, _RevealMetadataPreprocessor

from .base import ExportersTestsBase


class TestSlidesExporter(ExportersTestsBase):
    """Tests for SlidesExporter"""

    exporter_class = SlidesExporter  # type:ignore
    should_include_raw = ["html"]  # type:ignore

    def test_constructor(self):
        """
        Can a SlidesExporter be constructed?
        """
        SlidesExporter()

    def test_export(self):
        """
        Can a SlidesExporter export something?
        """
        (output, resources) = SlidesExporter().from_filename(self._get_notebook())
        assert len(output) > 0

    def test_export_reveal(self):
        """
        Can a SlidesExporter export using the 'reveal' template?
        """
        (output, resources) = SlidesExporter().from_filename(self._get_notebook())
        assert len(output) > 0

    def build_notebook(self):
        """Build a reveal slides notebook in memory for use with tests."""
        outputs = [nbformat.new_output(output_type="stream", name="stdout", text="a")]

        slide_metadata = {"slideshow": {"slide_type": "slide"}}
        subslide_metadata = {"slideshow": {"slide_type": "subslide"}}
        fragment_metadata = {"slideshow": {"slide_type": "fragment"}}

        cells = [
            nbformat.new_code_cell(source="", execution_count=1, outputs=outputs),
            nbformat.new_markdown_cell(source="", metadata=slide_metadata),
            nbformat.new_code_cell(source="", execution_count=2, outputs=outputs),
            nbformat.new_markdown_cell(source="", metadata=slide_metadata),
            nbformat.new_markdown_cell(source="", metadata=subslide_metadata),
            nbformat.new_markdown_cell(source="", metadata=fragment_metadata),
            nbformat.new_code_cell(source="", execution_count=1, outputs=outputs),
        ]

        return nbformat.new_notebook(cells=cells)

    def test_metadata_preprocessor(self):
        preprocessor = _RevealMetadataPreprocessor()
        nb = self.build_notebook()
        nb, resources = preprocessor.preprocess(nb)
        cells = nb.cells

        # Make sure correct metadata tags are available on every cell.
        for cell in cells:
            assert "slide_type" in cell.metadata

        # Make sure slide end is only applied to the cells preceding slide
        # cells.
        assert not cells[1].metadata.get("slide_end", False)

        # Verify 'slide-end'
        assert cells[0].metadata["slide_end"]
        assert cells[2].metadata["slide_end"]
        assert cells[2].metadata["subslide_end"]

        assert not cells[3].metadata.get("slide_end", False)
        assert cells[3].metadata["subslide_end"]

        assert cells[-1].metadata["fragment_end"]
        assert cells[-1].metadata["subslide_end"]
        assert cells[-1].metadata["slide_end"]



"""
Module with tests for Pandoc filters
"""

# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

import json

from nbconvert.filters.pandoc import ConvertExplicitlyRelativePaths, convert_pandoc
from tests.base import TestsBase
from tests.testutils import onlyif_cmds_exist


class TestPandocFilters(TestsBase):
    @onlyif_cmds_exist("pandoc")
    def test_convert_explicitly_relative_paths(self):
        """
        Do the image links in a markdown file located in dir get processed correctly?
        """
        inp_dir = "/home/user/src"
        fltr = ConvertExplicitlyRelativePaths(texinputs=inp_dir)

        # pairs of input, expected
        tests = {
            # TEXINPUTS is enough, abs_path not needed
            "im.png": "im.png",
            "./im.png": "im.png",
            "./images/im.png": "images/im.png",
            # TEXINPUTS is not enough, abs_path needed
            "../im.png": "/home/user/im.png",
            "../images/im.png": "/home/user/images/im.png",
            "../../images/im.png": "/home/images/im.png",
        }

        # this shouldn't be modified by the filter
        # since it is a code block inside markdown,
        # not an image link itself
        fake = """
        ```
        \\includegraphics{../fake.png}
        ```
        """

        # convert to markdown image
        def foo(filename):
            return f"![]({filename})"

        # create input markdown and convert to pandoc json
        inp = convert_pandoc(fake + "\n".join(map(foo, tests.keys())), "markdown", "json")
        expected = convert_pandoc(fake + "\n".join(map(foo, tests.values())), "markdown", "json")
        # Do this to fix string formatting
        expected = json.dumps(json.loads(expected))
        self.assertEqual(expected, fltr(inp))

    def test_convert_explicitly_relative_paths_no_texinputs(self):
        # no texinputs should lead to just returning
        fltr = ConvertExplicitlyRelativePaths(texinputs="")
        self.assertEqual("test", fltr("test"))



"""Tests for conversions from markdown to other formats"""

# Copyright (c) IPython Development Team.
# Distributed under the terms of the Modified BSD License.

import re
from copy import copy
from functools import partial
from html import unescape

from jinja2 import Environment

from nbconvert.filters.markdown import markdown2html
from nbconvert.filters.pandoc import convert_pandoc
from tests.base import TestsBase
from tests.testutils import onlyif_cmds_exist


class TestMarkdown(TestsBase):
    tests = [
        "*test",
        "**test",
        "*test*",
        "_test_",
        "__test__",
        "__*test*__",
        "**test**",
        "#test",
        "##test",
        "test\n----",
        "test [link](https://google.com/)",
    ]

    tokens = [
        "*test",
        "**test",
        "test",
        "test",
        "test",
        "test",
        "test",
        "test",
        "test",
        "test",
        ("test", "https://google.com/"),
    ]

    @onlyif_cmds_exist("pandoc")
    def test_markdown2latex(self):
        """markdown2latex test"""
        for index, test in enumerate(self.tests):
            self._try_markdown(
                partial(convert_pandoc, from_format="markdown", to_format="latex"),
                test,
                self.tokens[index],
            )

    @onlyif_cmds_exist("pandoc")
    def test_markdown2latex_markup(self):
        """markdown2latex with markup kwarg test"""
        # This string should be passed through unaltered with pandoc's
        # markdown_strict reader
        s = "1) arabic number with parenthesis"
        self.assertEqual(convert_pandoc(s, "markdown_strict", "latex"), s)
        # This string should be passed through unaltered with pandoc's
        # markdown_strict+tex_math_dollars reader
        s = r"$\alpha$ latex math"
        # sometimes pandoc uses $math$, sometimes it uses \(math\)
        expected = re.compile(r"(\$|\\\()\\alpha(\$|\\\)) latex math")

        assertRegex = self.assertRegex  # noqa

        assertRegex(convert_pandoc(s, "markdown_strict+tex_math_dollars", "latex"), expected)

    @onlyif_cmds_exist("pandoc")
    def test_pandoc_extra_args(self):
        # pass --no-wrap
        s = "\n".join(
            [
                "#latex {{long_line | md2l(['--wrap=none'])}}",
                "#rst {{long_line | md2r(['--columns', '5'])}}",
            ]
        )
        long_line = " ".join(["long"] * 30)
        env = Environment()  # noqa
        env.filters.update(
            {
                "md2l": lambda code, extra_args: convert_pandoc(
                    code, from_format="markdown", to_format="latex", extra_args=extra_args
                ),
                "md2r": lambda code, extra_args: convert_pandoc(
                    code, from_format="markdown", to_format="rst", extra_args=extra_args
                ),
            }
        )
        tpl = env.from_string(s)
        rendered = tpl.render(long_line=long_line)
        _, latex, rst = rendered.split("#")

        self.assertEqual(latex.strip(), "latex %s" % long_line)
        self.assertEqual(rst.strip(), "rst %s" % long_line.replace(" ", "\n"))

    def test_markdown2html(self):
        """markdown2html test"""
        for index, test in enumerate(self.tests):
            self._try_markdown(markdown2html, test, self.tokens[index])

    def test_markdown2html_heading_anchors(self):
        for md, tokens in [
            (
                "# test title",
                ("<h1", ">test title", 'id="test-title"', "&#182;</a>", "anchor-link"),
            ),
            (
                "###test head space",
                ("<h3", ">test head space", 'id="test-head-space"', "&#182;</a>", "anchor-link"),
            ),
        ]:
            self._try_markdown(markdown2html, md, tokens)

    def test_markdown2html_math(self):
        # Mathematical expressions not containing <, >, &
        # should be passed through unaltered
        # all the "<", ">", "&" must be escaped correctly
        cases = [
            (
                "\\begin{equation*}\n"
                + (
                    "\\left( \\sum_{k=1}^n a_k b_k \\right)^2 "
                    "\\leq \\left( \\sum_{k=1}^n a_k^2 \\right) "
                    "\\left( \\sum_{k=1}^n b_k^2 \\right)\n"
                )
                + "\\end{equation*}"
            ),
            ("$$\na = 1 *3* 5\n$$"),
            "$ a = 1 *3* 5 $",
            "$s_i = s_{i}\n$",
            "$a<b&b<lt$",
            "$a<b&lt;b>a;a-b<0$",
            "$<k'>$",
            "$$a<b&b<lt$$",
            "$$a<b&lt;b>a;a-b<0$$",
            "$$<k'>$$",
            ("$$x\n=\n2$$"),
            (
                "$$\n"
                "b =  \\left[\n"
                "P\\left(\\right)\n"
                "- (l_1\\leftrightarrow l_2\n)"
                "\\right]\n"
                "$$"
            ),
            ("\\begin{equation*}\nx = 2 *55* 7\n\\end{equation*}"),
            """$
\\begin{tabular}{ l c r }
  1 & 2 & 3 \\
  4 & 5 & 6 \\
  7 & 8 & 9 \\
\\end{tabular}$""",
        ]

        for case in cases:
            result = markdown2html(case)
            # find the equation in the generated texts
            search_result = re.search(r"\$.*\$", result, re.DOTALL)
            if search_result is None:
                search_result = re.search(
                    "\\\\begin\\{equation.*\\}.*\\\\end\\{equation.*\\}", result, re.DOTALL
                )
            math = search_result.group(0)  # type:ignore
            # the resulting math part can not contain "<", ">" or
            # "&" not followed by "lt;", "gt;", or "amp;".
            self.assertNotIn("<", math)
            self.assertNotIn(">", math)
            self.assertNotRegex(math, "&(?![gt;|lt;|amp;])")
            # the result should be able to be unescaped correctly
            self.assertEqual(case, unescape(math))

    def test_markdown2html_math_mixed(self):
        """ensure markdown between inline and inline-block math works and
        test multiple LaTeX markup syntaxes.
        """
        case = """The entries of \\\\(C\\\\) are given by the exact formula:
$$
C_{ik} = \\sum_{j=1}^n A_{ij} B_{jk},
$$
but you can _implement_ this computation in many ways.
$\approx 2mnp$ flops are needed for \\\\[ C_{ik} = \\sum_{j=1}^n A_{ij} B_{jk} \\\\].
Also check empty math blocks work correctly:
$$$$
\\\\[\\\\]"""
        output_check = (
            case.replace("_implement_", "<em>implement</em>")
            .replace("\\\\(", "$")
            .replace("\\\\)", "$")
            .replace("\\\\[", "$$")
            .replace("\\\\]", "$$")
        )
        # these replacements are needed because we use $ and $$ in our html output
        self._try_markdown(markdown2html, case, output_check)

    def test_markdown2html_math_paragraph(self):
        """these should all parse without modification"""
        cases = [
            # https://github.com/ipython/ipython/issues/6724
            """Water that is stored in $t$, $s_t$, must equal the storage content of the previous stage,
$s_{t-1}$, plus a stochastic inflow, $I_t$, minus what is being released in $t$, $r_t$.
With $s_0$ defined as the initial storage content in $t=1$, we have""",
            # https://github.com/jupyter/nbviewer/issues/420
            """$C_{ik}$
$$
C_{ik} = \\sum_{j=1}
$$
$C_{ik}$""",
            """$m$
$$
C = \begin{pmatrix}
          0 & 0 & 0 & \\cdots & 0 & 0 & -c_0 \\
          0 & 0 & 0 & \\cdots & 0 & 1 & -c_{m-1}
    \\end{pmatrix}
$$
$x^m$""",
            """$r=\\overline{1,n}$
$$ {\bf
b}_{i}^{r}(t)=(1-t)\\,{\bf b}_{i}^{r-1}(t)+t\\,{\bf b}_{i+1}^{r-1}(t),\\:
 i=\\overline{0,n-r}, $$
i.e. the $i^{th}$""",
        ]

        for case in cases:
            s = markdown2html(case)
            self.assertIn(case, unescape(s))

    @onlyif_cmds_exist("pandoc")
    def test_markdown2rst(self):
        """markdown2rst test"""

        # Modify token array for rst, escape asterisk
        tokens = copy(self.tokens)
        tokens[0] = r"\*test"
        tokens[1] = r"\**test"

        for index, test in enumerate(self.tests):
            self._try_markdown(
                partial(convert_pandoc, from_format="markdown", to_format="rst"),
                test,
                tokens[index],
            )

    def test_mermaid_markdown(self):
        code = """flowchart LR
            chicken --> egg --> chicken"""
        case = f"""```mermaid\n  {code}\n```"""

        output_check = (
            """<div class="jp-Mermaid"><pre class="mermaid">\n"""
            f"""{code.strip()}"""
            """\n</pre></div>"""
        )

        self._try_markdown(markdown2html, case, output_check)

    def _try_markdown(self, method, test, tokens):
        results = method(test)
        if isinstance(tokens, (str,)):
            self.assertIn(tokens, results)
        else:
            for token in tokens:
                self.assertIn(token, results)



"""Tests for the qtpdf preprocessor"""

# Copyright (c) IPython Development Team.
# Distributed under the terms of the Modified BSD License.

import pytest

from nbconvert.exporters.qt_screenshot import QT_INSTALLED
from nbconvert.exporters.qtpdf import QtPDFExporter

from .base import ExportersTestsBase


@pytest.mark.skipif(not QT_INSTALLED, reason="PyQtWebEngine not installed")
class TestQtPDFExporter(ExportersTestsBase):
    """Contains test functions for qtpdf.py"""

    exporter_class = QtPDFExporter  # type:ignore

    def test_export(self):
        """
        Can a TemplateExporter export something?
        """
        (output, resources) = QtPDFExporter().from_filename(self._get_notebook())
        assert len(output) > 0



"""Tests for PDF export"""

# Copyright (c) IPython Development Team.
# Distributed under the terms of the Modified BSD License.

import os
import shutil
from tempfile import TemporaryDirectory

from nbconvert.exporters.pdf import PDFExporter
from nbconvert.utils import _contextlib_chdir
from tests.testutils import onlyif_cmds_exist

from .base import ExportersTestsBase

# -----------------------------------------------------------------------------
# Class
# -----------------------------------------------------------------------------


class TestPDF(ExportersTestsBase):
    """Test PDF export"""

    exporter_class = PDFExporter  # type:ignore

    def test_constructor(self):
        """Can a PDFExporter be constructed?"""
        self.exporter_class()  # type:ignore

    @onlyif_cmds_exist("xelatex", "pandoc")
    def test_export(self):
        """Smoke test PDFExporter"""
        with TemporaryDirectory() as td:
            file_name = os.path.basename(self._get_notebook())
            newpath = os.path.join(td, file_name)
            shutil.copy(self._get_notebook(), newpath)
            (output, resources) = self.exporter_class(latex_count=1).from_filename(  # type:ignore
                newpath
            )
            self.assertIsInstance(output, bytes)
            assert len(output) > 0
            # all temporary file should be cleaned up
            assert {file_name} == set(os.listdir(td))

    @onlyif_cmds_exist("xelatex", "pandoc")
    def test_texinputs(self):
        """
        Is TEXINPUTS set properly when we are converting
        - in the same directory, and
        - in a different directory?
        """
        with TemporaryDirectory() as td, _contextlib_chdir.chdir(td):
            os.mkdir("folder")
            file_name = os.path.basename(self._get_notebook())
            nb1 = os.path.join(td, file_name)
            nb2 = os.path.join(td, "folder", file_name)
            ex1 = self.exporter_class(latex_count=1)  # type:ignore
            ex2 = self.exporter_class(latex_count=1)  # type:ignore
            shutil.copy(self._get_notebook(), nb1)
            shutil.copy(self._get_notebook(), nb2)
            _ = ex1.from_filename(nb1)
            _ = ex2.from_filename(nb2)
            assert ex1.texinputs == os.path.abspath(".")
            assert ex2.texinputs == os.path.abspath("./folder")



"""Tests for the ExtractAttachments preprocessor"""

# Copyright (c) IPython Development Team.
# Distributed under the terms of the Modified BSD License.

import os
from base64 import b64decode

from nbconvert.preprocessors.extractattachments import ExtractAttachmentsPreprocessor

from .base import PreprocessorTestsBase


class TestExtractAttachments(PreprocessorTestsBase):
    """Contains test functions for extractattachments.py"""

    def build_preprocessor(self):
        """Make an instance of a preprocessor"""
        preprocessor = ExtractAttachmentsPreprocessor()
        preprocessor.enabled = True
        return preprocessor

    def test_constructor(self):
        """Can a ExtractAttachmentsPreprocessor be constructed?"""
        self.build_preprocessor()

    def test_attachment(self):
        """Test the output of the ExtractAttachmentsPreprocessor"""
        nb = self.build_notebook(with_attachment=True)
        res = self.build_resources()
        preprocessor = self.build_preprocessor()
        nb, res = preprocessor(nb, res)

        # Check if attachment was extracted.
        attachments = nb.cells[-1].attachments
        self.assertIn("image.png", attachments)
        self.assertIn("image/png", attachments["image.png"])
        data = attachments["image.png"]["image/png"]
        # convert to bytes, b64 decode, convert to str
        data = b64decode(data.encode("utf-8"))
        self.assertEqual(data, b"test")

        # Verify attachment
        self.assertIn("image.png", res["outputs"])
        self.assertEqual(res["outputs"]["image.png"], b"test")

        # Verify cell source changed appropriately
        src = nb.cells[-1].source
        self.assertEqual(src, "![image.png](image.png)")

    def test_attachment_with_directory(self):
        """Test that cell source modifications are correct when files are put in a directory"""
        nb = self.build_notebook(with_attachment=True)
        res = self.build_resources()
        output_dir = "outputs"
        res["output_files_dir"] = output_dir
        preprocessor = self.build_preprocessor()
        nb, res = preprocessor(nb, res)

        # Verify attachment
        # This can have "\\" separator on Windows
        file_path = os.path.join("outputs", "image.png")
        self.assertIn(file_path, res["outputs"])

        # Verify cell source changed appropriately
        src = nb.cells[-1].source
        # This shouldn't change on Windows
        self.assertEqual(src, "![image.png](outputs/image.png)")

    def test_use_separate_dir_config(self):
        """Test that use_separate_dir and attachment_directory_template work properly"""
        nb = self.build_notebook(with_attachment=True)
        res = self.build_resources()
        res["unique_key"] = "notebook1"  # add notebook name for the folder
        preprocessor = self.build_preprocessor()
        preprocessor.use_separate_dir = True
        preprocessor.attachments_directory_template = "{notebook_name}_custom"
        nb, res = preprocessor(nb, res)

        # Verify attachment
        # This can have "\\" separator on Windows
        file_path = os.path.join("notebook1_custom", "image.png")
        self.assertIn(file_path, res["attachments"])

        # Verify cell source changed appropriately
        src = nb.cells[-1].source
        # This shouldn't change on Windows
        self.assertEqual(src, "![image.png](notebook1_custom/image.png)")



from .base import PostProcessorBase

# protect against unavailable tornado
try:
    from .serve import ServePostProcessor
except ImportError:
    ServePostProcessor = None  # type:ignore[misc,assignment]

__all__ = ["PostProcessorBase", "ServePostProcessor"]



"""utility functions for preprocessor tests"""

# Copyright (c) IPython Development Team.
# Distributed under the terms of the Modified BSD License.

from base64 import b64encode

from nbformat import v4 as nbformat

from nbconvert.exporters.exporter import ResourcesDict
from tests.base import TestsBase


class PreprocessorTestsBase(TestsBase):
    """Contains test functions preprocessor tests"""

    def build_notebook(self, with_json_outputs=False, with_attachment=False):
        """Build a notebook in memory for use with preprocessor tests"""

        outputs = [
            nbformat.new_output("stream", name="stdout", text="a"),
            nbformat.new_output("display_data", data={"text/plain": "b"}),
            nbformat.new_output("stream", name="stdout", text="c"),
            nbformat.new_output("stream", name="stdout", text="d"),
            nbformat.new_output("stream", name="stderr", text="e"),
            nbformat.new_output("stream", name="stderr", text="f"),
            nbformat.new_output("display_data", data={"image/png": "Zw=="}),  # g
            nbformat.new_output("display_data", data={"application/pdf": "aA=="}),  # h
        ]
        if with_json_outputs:
            outputs.extend(
                [
                    nbformat.new_output("display_data", data={"application/json": [1, 2, 3]}),  # j
                    nbformat.new_output(
                        "display_data", data={"application/json": {"a": 1, "c": {"b": 2}}}
                    ),  # k
                    nbformat.new_output("display_data", data={"application/json": "abc"}),  # l
                    nbformat.new_output("display_data", data={"application/json": 15.03}),  # m
                ]
            )

        cells = [
            nbformat.new_code_cell(source="$ e $", execution_count=1, outputs=outputs),
            nbformat.new_markdown_cell(source="$ e $"),
        ]

        if with_attachment:
            data = b"test"
            encoded_data = b64encode(data)
            # this is conversion of bytes to string, not base64 decoding
            attachments = {"image.png": {"image/png": encoded_data.decode()}}
            cells.extend(
                [
                    nbformat.new_markdown_cell(
                        source="![image.png](attachment:image.png)", attachments=attachments
                    )
                ]
            )

        return nbformat.new_notebook(cells=cells)

    def build_resources(self):
        """Build an empty resources dictionary."""

        res = ResourcesDict()
        res["metadata"] = ResourcesDict()
        return res



# -----------------------------------------------------------------------------
# Copyright (c) 2013, the IPython Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
# -----------------------------------------------------------------------------

import pytest

# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------
from nbconvert.filters.citation import citation2latex

# -----------------------------------------------------------------------------
# Tests
# -----------------------------------------------------------------------------
test_md = {
    """
# My Heading

Lorem ipsum dolor sit amet, consectetur adipiscing elit. Phasellus ac magna non augue
porttitor scelerisque ac id diam <cite data-cite="granger">Granger</cite>. Mauris elit
velit, lobortis sed interdum at, vestibulum vitae libero <strong data-cite="fperez">Perez</strong>.
Lorem ipsum dolor sit amet, consectetur adipiscing elit
<em data-cite="takluyver">Thomas</em>. Quisque iaculis ligula ut ipsum mattis viverra.

<p>Here is a plain paragraph that should be unaffected. It contains simple
relations like 1<2 & 4>5.</p>

* One <cite data-cite="jdfreder">Jonathan</cite>.
* Two <cite data-cite="carreau">Matthias</cite>.
* Three <cite data-cite="ivanov">Paul</cite>.
""": r"""
# My Heading

Lorem ipsum dolor sit amet, consectetur adipiscing elit. Phasellus ac magna non augue
porttitor scelerisque ac id diam \cite{granger}. Mauris elit
velit, lobortis sed interdum at, vestibulum vitae libero \cite{fperez}.
Lorem ipsum dolor sit amet, consectetur adipiscing elit
\cite{takluyver}. Quisque iaculis ligula ut ipsum mattis viverra.

<p>Here is a plain paragraph that should be unaffected. It contains simple
relations like 1<2 & 4>5.</p>

* One \cite{jdfreder}.
* Two \cite{carreau}.
* Three \cite{ivanov}.
""",
    # No citations
    r"""The quick brown fox jumps over the lazy dog.""": r"""The quick brown fox jumps over the lazy dog.""",
    # Simple inline
    r"""Foo <cite data-cite=asdf>Text</cite> bar""": r"""Foo \cite{asdf} bar""",
    # Multiline
    r"""<cite data-cite=ewqr>Text
</cite>Foo""": r"""\cite{ewqr}Foo""",
    # Nested tags
    r"""<div><div data-cite=Foo><div>Text</div></div></div> Bar""": r"""<div>\cite{Foo}</div> Bar""",
    # Including Maths
    r"""Foo $3*2*1$ <div data-cite=Foo>Text</div> Bar""": r"""Foo $3*2*1$ \cite{Foo} Bar""",
    # Missing end tag
    r"""<cite data-cite=asdf>Test Foo""": r"""\cite{asdf}""",
    r"""<cite data-cite=asdf><cite>Test Foo""": r"""\cite{asdf}""",
    r"""<cite data-cite=asdf><cite>Test</cite> Foo""": r"""\cite{asdf}""",
    # Multiple arguments
    r"""<cite width=qwer data-cite=asdf>Test</cite> Foo""": r"""\cite{asdf} Foo""",
    # Wrong capitalization
    r"""<CITE data-cite=asdf>Test</cite> Foo""": r"""\cite{asdf} Foo""",
    r"""<cite DATA-CITE=asdf>Test</cite> Foo""": r"""\cite{asdf} Foo""",
    # Wrong end tag
    r"""<asd data-cite=wer> ksjfs </asdf> sdf ds """: r"""\cite{wer}""",
    r"""<asd data-cite=wer>""": r"""\cite{wer}""",
    # Invalid tag names
    r"""<frog> <foo data-cite=wer></foo>""": r"""<frog> \cite{wer}""",
    # Non-nested tags
    r"""<strong> <h1> <cite data-cite=asdf></cite>Test</strong> Foo </h1>""": r"""<strong> <h1> \cite{asdf}Test</strong> Foo </h1>""",
    # LXML errors
    r"""Foo
\begin{eqnarray}
1 & <cite data-cite=bar>bar1</cite> \\
3 & 4 \\
\end{eqnarray}""": r"""Foo
\begin{eqnarray}
1 & \cite{bar} \\
3 & 4 \\
\end{eqnarray}""",
    r"""
1<2 is true, but 3>4 is false.

$1<2$ is true, but $3>4$ is false.

1<2 it is even worse if it is alone in a line.""": r"""
1<2 is true, but 3>4 is false.

$1<2$ is true, but $3>4$ is false.

1<2 it is even worse if it is alone in a line.""",
    r"""
1 < 2 is true, but 3 > 4 is false

$1 < 2$ is true, but $3 > 4$ is false

1 < 2 it is even worse if it is alone in a line.
""": r"""
1 < 2 is true, but 3 > 4 is false

$1 < 2$ is true, but $3 > 4$ is false

1 < 2 it is even worse if it is alone in a line.
""",
}


@pytest.mark.parametrize(["in_arg", "out_arg"], list(test_md.items()))
def test_citation2latex(in_arg, out_arg):
    """Are citations parsed properly?"""
    assert citation2latex(in_arg) == out_arg



"""Markdown filters with mistune

Used from markdown.py
"""
# Copyright (c) IPython Development Team.
# Distributed under the terms of the Modified BSD License.


import base64
import mimetypes
import os
from html import escape
from typing import Any, Callable, Dict, Iterable, Match, Optional, Tuple

import bs4
from pygments import highlight
from pygments.formatters import HtmlFormatter
from pygments.lexer import Lexer
from pygments.lexers import get_lexer_by_name
from pygments.util import ClassNotFound

from nbconvert.filters.strings import add_anchor

try:  # for Mistune >= 3.0
    from mistune import (  # type:ignore[attr-defined]
        BlockParser,
        BlockState,
        HTMLRenderer,
        InlineParser,
        InlineState,
        Markdown,
        import_plugin,
    )

    MISTUNE_V3 = True

except ImportError:  # for Mistune >= 2.0
    import re

    from mistune import (  # type: ignore[attr-defined]
        PLUGINS,
        BlockParser,
        HTMLRenderer,
        InlineParser,
        Markdown,
    )

    MISTUNE_V3 = False

    def import_plugin(name: str) -> "MarkdownPlugin":  # type: ignore[misc]
        """Simple implementation of Mistune V3's import_plugin for V2."""
        return PLUGINS[name]  # type: ignore[no-any-return]


class InvalidNotebook(Exception):  # noqa
    """An invalid notebook model."""

    pass


def _dotall(pattern: str) -> str:
    """Makes the '.' special character match any character inside the pattern, including a newline.

    This is implemented with the inline flag `(?s:...)` and is equivalent to using `re.DOTALL`.
    It is useful for LaTeX environments, where line breaks may be present.
    """
    return f"(?s:{pattern})"


if MISTUNE_V3:  # Parsers for Mistune >= 3.0.0

    class MathBlockParser(BlockParser):
        """This acts as a pass-through to the MathInlineParser. It is needed in
        order to avoid other block level rules splitting math sections apart.

        It works by matching each multiline math environment as a single paragraph,
        so that other rules don't think each section is its own paragraph. Inline
        is ignored here.
        """

        AXT_HEADING_WITHOUT_LEADING_SPACES = (
            r"^ {0,3}(?P<axt_1>#{1,6})(?!#+)(?P<axt_2>[ \t]*(.*?)?)$"
        )

        MULTILINE_MATH = _dotall(
            # Display math mode, old TeX delimiter: $$ \sqrt{2} $$
            r"(?<!\\)[$]{2}.*?(?<!\\)[$]{2}"
            "|"
            # Display math mode, new LaTeX delimiter: \[ \sqrt{2} \]
            r"\\\\\[.*?\\\\\]"
            "|"
            # LaTeX environment: \begin{equation} \sqrt{2} \end{equation}
            r"\\begin\{(?P<math_env_name>[a-z]*\*?)\}.*?\\end\{(?P=math_env_name)\}"
        )

        SPECIFICATION = {
            **BlockParser.SPECIFICATION,
            "axt_heading": AXT_HEADING_WITHOUT_LEADING_SPACES,
            "multiline_math": MULTILINE_MATH,
        }

        # Multiline math must be searched before other rules
        DEFAULT_RULES: Tuple[str, ...] = ("multiline_math", *BlockParser.DEFAULT_RULES)  # type: ignore[assignment]

        def parse_multiline_math(self, m: Match[str], state: BlockState) -> int:
            """Send mutiline math as a single paragraph to MathInlineParser."""
            matched_text = m[0]
            state.add_paragraph(matched_text)
            return m.end()

    class MathInlineParser(InlineParser):
        r"""This interprets the content of LaTeX style math objects.

        In particular this grabs ``$$...$$``, ``\\[...\\]``, ``\\(...\\)``, ``$...$``,
        and ``\begin{foo}...\end{foo}`` styles for declaring mathematics. It strips
        delimiters from all these varieties, and extracts the type of environment
        in the last case (``foo`` in this example).
        """

        # Display math mode, using older TeX delimiter: $$ \pi $$
        BLOCK_MATH_TEX = _dotall(r"(?<!\\)\$\$(?P<math_block_tex>.*?)(?<!\\)\$\$")
        # Display math mode, using newer LaTeX delimiter: \[ \pi \]
        BLOCK_MATH_LATEX = _dotall(r"(?<!\\)\\\\\[(?P<math_block_latex>.*?)(?<!\\)\\\\\]")
        # Inline math mode, using older TeX delimiter: $ \pi $  (cannot be empty!)
        INLINE_MATH_TEX = _dotall(r"(?<![$\\])\$(?P<math_inline_tex>.+?)(?<![$\\])\$")
        # Inline math mode, using newer LaTeX delimiter: \( \pi \)
        INLINE_MATH_LATEX = _dotall(r"(?<!\\)\\\\\((?P<math_inline_latex>.*?)(?<!\\)\\\\\)")
        # LaTeX math environment: \begin{equation} \pi \end{equation}
        LATEX_ENVIRONMENT = _dotall(
            r"\\begin\{(?P<math_env_name>[a-z]*\*?)\}"
            r"(?P<math_env_body>.*?)"
            r"\\end\{(?P=math_env_name)\}"
        )

        SPECIFICATION = {
            **InlineParser.SPECIFICATION,
            "block_math_tex": BLOCK_MATH_TEX,
            "block_math_latex": BLOCK_MATH_LATEX,
            "inline_math_tex": INLINE_MATH_TEX,
            "inline_math_latex": INLINE_MATH_LATEX,
            "latex_environment": LATEX_ENVIRONMENT,
        }

        # Block math must be matched first, and all math must come before text
        DEFAULT_RULES: Tuple[str, ...] = (
            "block_math_tex",
            "block_math_latex",
            "inline_math_tex",
            "inline_math_latex",
            "latex_environment",
            *InlineParser.DEFAULT_RULES,
        )  # type: ignore[assignment]

        def parse_block_math_tex(self, m: Match[str], state: InlineState) -> int:
            """Parse older TeX-style display math."""
            body = m.group("math_block_tex")
            state.append_token({"type": "block_math", "raw": body})
            return m.end()

        def parse_block_math_latex(self, m: Match[str], state: InlineState) -> int:
            """Parse newer LaTeX-style display math."""
            body = m.group("math_block_latex")
            state.append_token({"type": "block_math", "raw": body})
            return m.end()

        def parse_inline_math_tex(self, m: Match[str], state: InlineState) -> int:
            """Parse older TeX-style inline math."""
            body = m.group("math_inline_tex")
            state.append_token({"type": "inline_math", "raw": body})
            return m.end()

        def parse_inline_math_latex(self, m: Match[str], state: InlineState) -> int:
            """Parse newer LaTeX-style inline math."""
            body = m.group("math_inline_latex")
            state.append_token({"type": "inline_math", "raw": body})
            return m.end()

        def parse_latex_environment(self, m: Match[str], state: InlineState) -> int:
            """Parse a latex environment."""
            attrs = {"name": m.group("math_env_name"), "body": m.group("math_env_body")}
            state.append_token({"type": "latex_environment", "attrs": attrs})
            return m.end()

else:  # Parsers for Mistune >= 2.0.0 < 3.0.0

    class MathBlockParser(BlockParser):  # type: ignore[no-redef]
        """This acts as a pass-through to the MathInlineParser. It is needed in
        order to avoid other block level rules splitting math sections apart.
        """

        MULTILINE_MATH = re.compile(
            # Display math mode, old TeX delimiter: $$ \sqrt{2} $$
            r"(?<!\\)[$]{2}.*?(?<!\\)[$]{2}|"
            # Display math mode, new LaTeX delimiter: \[ \sqrt{2} \]
            r"\\\\\[.*?\\\\\]|"
            # LaTeX environment: \begin{equation} \sqrt{2} \end{equation}
            r"\\begin\{([a-z]*\*?)\}.*?\\end\{\1\}",
            re.DOTALL,
        )

        # Regex for header that doesn't require space after '#'
        AXT_HEADING = re.compile(r" {0,3}(#{1,6})(?!#+)(?: *\n+|([^\n]*?)(?:\n+|\s+?#+\s*\n+))")

        # Multiline math must be searched before other rules
        RULE_NAMES = ("multiline_math", *BlockParser.RULE_NAMES)  # type: ignore[attr-defined]

        def parse_multiline_math(self, m: Match[str], state: Any) -> Dict[str, str]:
            """Pass token through mutiline math."""
            return {"type": "multiline_math", "text": m.group(0)}

    class MathInlineParser(InlineParser):  # type: ignore[no-redef]
        r"""This interprets the content of LaTeX style math objects.

        In particular this grabs ``$$...$$``, ``\\[...\\]``, ``\\(...\\)``, ``$...$``,
        and ``\begin{foo}...\end{foo}`` styles for declaring mathematics. It strips
        delimiters from all these varieties, and extracts the type of environment
        in the last case (``foo`` in this example).
        """

        # Display math mode, using older TeX delimiter: $$ \pi $$
        BLOCK_MATH_TEX = _dotall(r"(?<!\\)\$\$(.*?)(?<!\\)\$\$")
        # Display math mode, using newer LaTeX delimiter: \[ \pi \]
        BLOCK_MATH_LATEX = _dotall(r"(?<!\\)\\\\\[(.*?)(?<!\\)\\\\\]")
        # Inline math mode, using older TeX delimiter: $ \pi $  (cannot be empty!)
        INLINE_MATH_TEX = _dotall(r"(?<![$\\])\$(.+?)(?<![$\\])\$")
        # Inline math mode, using newer LaTeX delimiter: \( \pi \)
        INLINE_MATH_LATEX = _dotall(r"(?<!\\)\\\\\((.*?)(?<!\\)\\\\\)")
        # LaTeX math environment: \begin{equation} \pi \end{equation}
        LATEX_ENVIRONMENT = _dotall(r"\\begin\{([a-z]*\*?)\}(.*?)\\end\{\1\}")

        RULE_NAMES = (
            "block_math_tex",
            "block_math_latex",
            "inline_math_tex",
            "inline_math_latex",
            "latex_environment",
            *InlineParser.RULE_NAMES,  # type: ignore[attr-defined]
        )

        def parse_block_math_tex(self, m: Match[str], state: Any) -> Tuple[str, str]:
            """Parse block text math."""
            # sometimes the Scanner keeps the final '$$', so we use the
            # full matched string and remove the math markers
            text = m.group(0)[2:-2]
            return "block_math", text

        def parse_block_math_latex(self, m: Match[str], state: Any) -> Tuple[str, str]:
            """Parse block latex math ."""
            text = m.group(1)
            return "block_math", text

        def parse_inline_math_tex(self, m: Match[str], state: Any) -> Tuple[str, str]:
            """Parse inline tex math."""
            text = m.group(1)
            return "inline_math", text

        def parse_inline_math_latex(self, m: Match[str], state: Any) -> Tuple[str, str]:
            """Parse inline latex math."""
            text = m.group(1)
            return "inline_math", text

        def parse_latex_environment(self, m: Match[str], state: Any) -> Tuple[str, str, str]:
            """Parse a latex environment."""
            name, text = m.group(1), m.group(2)
            return "latex_environment", name, text


class IPythonRenderer(HTMLRenderer):
    """An ipython html renderer."""

    def __init__(  # noqa
        self,
        escape: bool = True,
        allow_harmful_protocols: bool = True,
        embed_images: bool = False,
        exclude_anchor_links: bool = False,
        anchor_link_text: str = "¶",
        path: str = "",
        attachments: Optional[Dict[str, Dict[str, str]]] = None,
    ):
        """Initialize the renderer."""
        super().__init__(escape, allow_harmful_protocols)
        self.embed_images = embed_images
        self.exclude_anchor_links = exclude_anchor_links
        self.anchor_link_text = anchor_link_text
        self.path = path
        if attachments is not None:
            self.attachments = attachments
        else:
            self.attachments = {}

    def block_code(self, code: str, info: Optional[str] = None) -> str:
        """Handle block code."""
        lang: Optional[str] = ""
        lexer: Optional[Lexer] = None

        if info:
            if info.startswith("mermaid"):
                return self.block_mermaidjs(code)

            try:
                if info.strip().split(None, 1):
                    lang = info.strip().split(maxsplit=1)[0]
                    lexer = get_lexer_by_name(lang, stripall=True)
            except ClassNotFound:
                code = f"{lang}\n{code}"
                lang = None

        if not lang:
            return super().block_code(code, info=info)

        formatter = HtmlFormatter()
        return highlight(code, lexer, formatter)

    def block_mermaidjs(self, code: str) -> str:
        """Handle mermaid syntax."""
        return (
            """<div class="jp-Mermaid"><pre class="mermaid">\n"""
            f"""{code.strip()}"""
            """\n</pre></div>"""
        )

    def block_html(self, html: str) -> str:
        """Handle block html."""
        if self.embed_images:
            html = self._html_embed_images(html)

        return super().block_html(html)

    def inline_html(self, html: str) -> str:
        """Handle inline html."""
        if self.embed_images:
            html = self._html_embed_images(html)

        return super().inline_html(html)

    def heading(self, text: str, level: int, **attrs: Dict[str, Any]) -> str:
        """Handle a heading."""
        html = super().heading(text, level, **attrs)
        if self.exclude_anchor_links:
            return html
        return str(add_anchor(html, anchor_link_text=self.anchor_link_text))

    def escape_html(self, text: str) -> str:
        """Escape html content."""
        return escape(text, quote=False)

    def block_math(self, body: str) -> str:
        """Handle block math."""
        return f"$${self.escape_html(body)}$$"

    def multiline_math(self, text: str) -> str:
        """Handle mulitline math for older mistune versions."""
        return text

    def latex_environment(self, name: str, body: str) -> str:
        """Handle a latex environment."""
        name, body = self.escape_html(name), self.escape_html(body)
        return f"\\begin{{{name}}}{body}\\end{{{name}}}"

    def inline_math(self, body: str) -> str:
        """Handle inline math."""
        return f"${self.escape_html(body)}$"

    def image(self, text: str, url: str, title: Optional[str] = None) -> str:
        """Rendering a image with title and text.

        :param text: alt text of the image.
        :param url: source link of the image.
        :param title: title text of the image.

        :note: The parameters `text` and `url` are swapped in older versions
            of mistune.
        """
        if MISTUNE_V3:
            url = self._embed_image_or_attachment(url)
        else:  # for mistune v2, the first argument is the URL
            text = self._embed_image_or_attachment(text)

        return super().image(text, url, title)

    def _embed_image_or_attachment(self, src: str) -> str:
        """Embed an image or attachment, depending on the configuration.
        If neither is possible, returns the original URL.
        """

        attachment_prefix = "attachment:"
        if src.startswith(attachment_prefix):
            name = src[len(attachment_prefix) :]

            if name not in self.attachments:
                msg = f"missing attachment: {name}"
                raise InvalidNotebook(msg)

            attachment = self.attachments[name]
            # we choose vector over raster, and lossless over lossy
            preferred_mime_types = ("image/svg+xml", "image/png", "image/jpeg")
            for mime_type in preferred_mime_types:
                if mime_type in attachment:
                    return f"data:{mime_type};base64,{attachment[mime_type]}"
            # otherwise we choose the first mimetype we can find
            default_mime_type = next(iter(attachment.keys()))
            return f"data:{default_mime_type};base64,{attachment[default_mime_type]}"

        elif self.embed_images:
            base64_url = self._src_to_base64(src)
            if base64_url is not None:
                return base64_url

        return src

    def _src_to_base64(self, src: str) -> Optional[str]:
        """Turn the source file into a base64 url.

        :param src: source link of the file.
        :return: the base64 url or None if the file was not found.
        """
        src_path = os.path.join(self.path, src)

        if not os.path.exists(src_path):
            return None

        with open(src_path, "rb") as fobj:
            mime_type, _ = mimetypes.guess_type(src_path)

            base64_data = base64.b64encode(fobj.read())
            base64_str = base64_data.replace(b"\n", b"").decode("ascii")

            return f"data:{mime_type};base64,{base64_str}"

    def _html_embed_images(self, html: str) -> str:
        parsed_html = bs4.BeautifulSoup(html, features="html.parser")
        imgs: bs4.ResultSet[bs4.Tag] = parsed_html.find_all("img")

        # Replace img tags's sources by base64 dataurls
        for img in imgs:
            src = img.attrs.get("src")
            if src is None:
                continue

            base64_url = self._src_to_base64(img.attrs["src"])
            if base64_url is not None:
                img.attrs["src"] = base64_url

        return str(parsed_html)


# Represents an already imported plugin for Mistune
MarkdownPlugin = Callable[[Markdown], None]


class MarkdownWithMath(Markdown):
    """Markdown text with math enabled."""

    DEFAULT_PLUGINS = (
        # "abbr",  (see https://github.com/jupyter/nbconvert/pull/1853)
        # "footnotes",
        "strikethrough",
        "table",
        "url",
        "task_lists",
        "def_list",
    )

    def __init__(
        self,
        renderer: HTMLRenderer,
        block: Optional[BlockParser] = None,
        inline: Optional[InlineParser] = None,
        plugins: Optional[Iterable[MarkdownPlugin]] = None,
    ):
        """Initialize the parser."""
        if block is None:
            block = MathBlockParser()
        if inline is None:
            if MISTUNE_V3:
                inline = MathInlineParser(hard_wrap=False)
            else:
                inline = MathInlineParser(renderer, hard_wrap=False)  # type: ignore[arg-type,misc]
        if plugins is None:
            plugins = (import_plugin(p) for p in self.DEFAULT_PLUGINS)

        super().__init__(renderer, block, inline, plugins)

    def render(self, source: str) -> str:
        """Render the HTML output for a Markdown source."""
        return str(super().__call__(source))


def markdown2html_mistune(source: str) -> str:
    """Convert a markdown string to HTML using mistune"""
    return MarkdownWithMath(renderer=IPythonRenderer(escape=False)).render(source)



"""Filters for processing ANSI colors within Jinja templates."""

# Copyright (c) IPython Development Team.
# Distributed under the terms of the Modified BSD License.

import re

import markupsafe

__all__ = ["strip_ansi", "ansi2html", "ansi2latex"]

_ANSI_RE = re.compile("\x1b\\[(.*?)([@-~])")

_ANSI_COLORS = (
    "ansi-black",
    "ansi-red",
    "ansi-green",
    "ansi-yellow",
    "ansi-blue",
    "ansi-magenta",
    "ansi-cyan",
    "ansi-white",
    "ansi-black-intense",
    "ansi-red-intense",
    "ansi-green-intense",
    "ansi-yellow-intense",
    "ansi-blue-intense",
    "ansi-magenta-intense",
    "ansi-cyan-intense",
    "ansi-white-intense",
)


def strip_ansi(source):
    """
    Remove ANSI escape codes from text.

    Parameters
    ----------
    source : str
        Source to remove the ANSI from

    """
    return _ANSI_RE.sub("", source)


def ansi2html(text):
    """
    Convert ANSI colors to HTML colors.

    Parameters
    ----------
    text : unicode
        Text containing ANSI colors to convert to HTML

    """
    text = markupsafe.escape(text)
    return _ansi2anything(text, _htmlconverter)


def ansi2latex(text):
    """
    Convert ANSI colors to LaTeX colors.

    Parameters
    ----------
    text : unicode
        Text containing ANSI colors to convert to LaTeX

    """
    return _ansi2anything(text, _latexconverter)


def _htmlconverter(fg, bg, bold, underline, inverse):  # noqa
    """
    Return start and end tags for given foreground/background/bold/underline.

    """
    if (fg, bg, bold, underline, inverse) == (None, None, False, False, False):
        return "", ""

    classes = []
    styles = []

    if inverse:
        fg, bg = bg, fg

    if isinstance(fg, int):
        classes.append(_ANSI_COLORS[fg] + "-fg")
    elif fg:
        styles.append("color: rgb({},{},{})".format(*fg))
    elif inverse:
        classes.append("ansi-default-inverse-fg")

    if isinstance(bg, int):
        classes.append(_ANSI_COLORS[bg] + "-bg")
    elif bg:
        styles.append("background-color: rgb({},{},{})".format(*bg))
    elif inverse:
        classes.append("ansi-default-inverse-bg")

    if bold:
        classes.append("ansi-bold")

    if underline:
        classes.append("ansi-underline")

    starttag = "<span"
    if classes:
        starttag += ' class="' + " ".join(classes) + '"'
    if styles:
        starttag += ' style="' + "; ".join(styles) + '"'
    starttag += ">"
    return starttag, "</span>"


def _latexconverter(fg, bg, bold, underline, inverse):  # noqa
    """
    Return start and end markup given foreground/background/bold/underline.

    """
    if (fg, bg, bold, underline, inverse) == (None, None, False, False, False):
        return "", ""

    starttag, endtag = "", ""

    if inverse:
        fg, bg = bg, fg

    if isinstance(fg, int):
        starttag += r"\textcolor{" + _ANSI_COLORS[fg] + "}{"
        endtag = "}" + endtag
    elif fg:
        # See http://tex.stackexchange.com/a/291102/13684
        starttag += r"\def\tcRGB{\textcolor[RGB]}\expandafter"
        starttag += r"\tcRGB\expandafter{{\detokenize{{{},{},{}}}}}{{".format(*fg)
        endtag = "}" + endtag
    elif inverse:
        starttag += r"\textcolor{ansi-default-inverse-fg}{"
        endtag = "}" + endtag

    if isinstance(bg, int):
        starttag += r"\setlength{\fboxsep}{0pt}"
        starttag += r"\colorbox{" + _ANSI_COLORS[bg] + "}{"
        endtag = r"\strut}" + endtag
    elif bg:
        starttag += r"\setlength{\fboxsep}{0pt}"
        # See http://tex.stackexchange.com/a/291102/13684
        starttag += r"\def\cbRGB{\colorbox[RGB]}\expandafter"
        starttag += r"\cbRGB\expandafter{{\detokenize{{{},{},{}}}}}{{".format(*bg)
        endtag = r"\strut}" + endtag
    elif inverse:
        starttag += r"\setlength{\fboxsep}{0pt}"
        starttag += r"\colorbox{ansi-default-inverse-bg}{"
        endtag = r"\strut}" + endtag

    if bold:
        starttag += r"\textbf{"
        endtag = "}" + endtag

    if underline:
        starttag += r"\underline{"
        endtag = "}" + endtag

    return starttag, endtag


def _ansi2anything(text, converter):  # noqa
    r"""
    Convert ANSI colors to HTML or LaTeX.

    See https://en.wikipedia.org/wiki/ANSI_escape_code

    Accepts codes like '\x1b[32m' (red) and '\x1b[1;32m' (bold, red).

    Non-color escape sequences (not ending with 'm') are filtered out.

    Ideally, this should have the same behavior as the function
    fixConsole() in notebook/notebook/static/base/js/utils.js.

    """
    fg, bg = None, None
    bold = False
    underline = False
    inverse = False
    numbers = []
    out = []

    while text:
        m = _ANSI_RE.search(text)
        if m:
            if m.group(2) == "m":
                try:
                    # Empty code is same as code 0
                    numbers = [int(n) if n else 0 for n in m.group(1).split(";")]
                except ValueError:
                    pass  # Invalid color specification
            else:
                pass  # Not a color code
            chunk, text = text[: m.start()], text[m.end() :]
        else:
            chunk, text = text, ""

        if chunk:
            starttag, endtag = converter(
                fg + 8 if bold and fg in range(8) else fg,  # type:ignore[operator]
                bg,
                bold,
                underline,
                inverse,
            )
            out.append(starttag)
            out.append(chunk)
            out.append(endtag)

        while numbers:
            n = numbers.pop(0)
            if n == 0:
                # Code 0 (same as empty code): reset everything
                fg = bg = None
                bold = underline = inverse = False
            elif n == 1:
                bold = True
            elif n == 4:
                underline = True
            elif n == 5:
                # Code 5: blinking
                bold = True
            elif n == 7:
                inverse = True
            elif n in (21, 22):
                bold = False
            elif n == 24:
                underline = False
            elif n == 27:
                inverse = False
            elif 30 <= n <= 37:
                fg = n - 30
            elif n == 38:
                try:
                    fg = _get_extended_color(numbers)
                except ValueError:
                    numbers.clear()
            elif n == 39:
                fg = None
            elif 40 <= n <= 47:
                bg = n - 40
            elif n == 48:
                try:
                    bg = _get_extended_color(numbers)
                except ValueError:
                    numbers.clear()
            elif n == 49:
                bg = None
            elif 90 <= n <= 97:
                fg = n - 90 + 8
            elif 100 <= n <= 107:
                bg = n - 100 + 8
            else:
                pass  # Unknown codes are ignored
    return "".join(out)


def _get_extended_color(numbers):
    n = numbers.pop(0)
    if n == 2 and len(numbers) >= 3:
        # 24-bit RGB
        r = numbers.pop(0)
        g = numbers.pop(0)
        b = numbers.pop(0)
        if not all(0 <= c <= 255 for c in (r, g, b)):
            raise ValueError()
    elif n == 5 and len(numbers) >= 1:
        # 256 colors
        idx = numbers.pop(0)
        if idx < 0:
            raise ValueError()
        elif idx < 16:
            # 16 default terminal colors
            return idx
        elif idx < 232:
            # 6x6x6 color cube, see http://stackoverflow.com/a/27165165/500098
            r = (idx - 16) // 36
            r = 55 + r * 40 if r > 0 else 0
            g = ((idx - 16) % 36) // 6
            g = 55 + g * 40 if g > 0 else 0
            b = (idx - 16) % 6
            b = 55 + b * 40 if b > 0 else 0
        elif idx < 256:
            # grayscale, see http://stackoverflow.com/a/27165165/500098
            r = g = b = (idx - 232) * 10 + 8
        else:
            raise ValueError()
    else:
        raise ValueError()
    return r, g, b



"""Tests for the svg2pdf preprocessor"""

# Copyright (c) IPython Development Team.
# Distributed under the terms of the Modified BSD License.

from unittest.mock import patch

from nbformat import v4 as nbformat

from nbconvert.preprocessors.svg2pdf import SVG2PDFPreprocessor
from tests.testutils import onlyif_cmds_exist

from .base import PreprocessorTestsBase


class Testsvg2pdf(PreprocessorTestsBase):
    """Contains test functions for svg2pdf.py"""

    simple_svg = """<?xml version="1.0" encoding="UTF-8" standalone="no"?>
<!-- Created with Inkscape (http://www.inkscape.org/) -->
<svg
   xmlns:svg="http://www.w3.org/2000/svg"
   xmlns="http://www.w3.org/2000/svg"
   version="1.0"
   x="0.00000000"
   y="0.00000000"
   width="500.00000"
   height="500.00000"
   id="svg2">
  <defs
     id="defs4" />
  <g
     id="layer1">
    <rect
       width="300.00000"
       height="300.00000"
       x="100.00000"
       y="100.00000"
       style="opacity:1.0000000;fill:none;fill-opacity:1.0000000;fill-rule:evenodd;stroke:#000000;stroke-width:8.0000000;stroke-linecap:round;stroke-linejoin:round;stroke-miterlimit:4.0000000;stroke-dasharray:none;stroke-dashoffset:0.00000000;stroke-opacity:1.0000000"
       id="rect5719" />
  </g>
</svg>"""

    def build_notebook(self):
        """Build a reveal slides notebook in memory for use with tests.
        Overrides base in PreprocessorTestsBase"""

        outputs = [
            nbformat.new_output(output_type="display_data", data={"image/svg+xml": self.simple_svg})
        ]

        cells = [nbformat.new_code_cell(source="", execution_count=1, outputs=outputs)]

        return nbformat.new_notebook(cells=cells)

    def build_preprocessor(self, **kwargs):
        """Make an instance of a preprocessor"""
        preprocessor = SVG2PDFPreprocessor(**kwargs)
        preprocessor.enabled = True
        return preprocessor

    def test_constructor(self):
        """Can a SVG2PDFPreprocessor be constructed?"""
        self.build_preprocessor()

    @onlyif_cmds_exist("inkscape")
    def test_output(self):
        """Test the output of the SVG2PDFPreprocessor"""
        nb = self.build_notebook()
        res = self.build_resources()
        preprocessor = self.build_preprocessor()
        nb, res = preprocessor(nb, res)
        self.assertIn("application/pdf", nb.cells[0].outputs[0].data)

    @onlyif_cmds_exist("inkscape")
    @patch("subprocess.Popen")
    def test_inkscape_version_default(self, mock_popen):
        mock_popen().communicate.return_value = (b"Inkscape 0.92.3 (2405546, 2018-03-11)", b"")
        mock_popen().returncode = 0

        preprocessor = self.build_preprocessor()
        assert preprocessor.inkscape_version == "0.92.3"

    def test_inkscape_pre_v1_command(self):
        preprocessor = self.build_preprocessor(inkscape="fake-inkscape", inkscape_version="0.92.3")
        assert preprocessor.command == [
            "fake-inkscape",
            "--without-gui",
            "--export-pdf={to_filename}",
            "{from_filename}",
        ]

    def test_inkscape_v1_command(self):
        preprocessor = self.build_preprocessor(
            inkscape="fake-inkscape", inkscape_version="1.0beta2"
        )
        assert preprocessor.command == [
            "fake-inkscape",
            "--export-filename={to_filename}",
            "{from_filename}",
        ]



"""Export to PDF via latex"""

# Copyright (c) IPython Development Team.
# Distributed under the terms of the Modified BSD License.
from __future__ import annotations

import os
import shutil
import subprocess
import sys
from tempfile import TemporaryDirectory

from traitlets import Bool, Instance, Integer, List, Unicode, default

from nbconvert.utils import _contextlib_chdir

from .latex import LatexExporter


class LatexFailed(IOError):  # noqa
    """Exception for failed latex run

    Captured latex output is in error.output.
    """

    def __init__(self, output):
        """Initialize the error."""
        self.output = output

    def __unicode__(self):
        """Unicode representation."""
        return "PDF creating failed, captured latex output:\n%s" % self.output

    def __str__(self):
        """String representation."""
        u = self.__unicode__()
        return u


def prepend_to_env_search_path(varname, value, envdict):
    """Add value to the environment variable varname in envdict

    e.g. prepend_to_env_search_path('BIBINPUTS', '/home/sally/foo', os.environ)
    """
    if not value:
        return  # Nothing to add

    envdict[varname] = value + os.pathsep + envdict.get(varname, "")


class PDFExporter(LatexExporter):
    """Writer designed to write to PDF files.

    This inherits from `LatexExporter`. It creates a LaTeX file in
    a temporary directory using the template machinery, and then runs LaTeX
    to create a pdf.
    """

    export_from_notebook = "PDF via LaTeX"

    latex_count = Integer(3, help="How many times latex will be called.").tag(config=True)

    latex_command = List(
        ["xelatex", "{filename}", "-quiet"], help="Shell command used to compile latex."
    ).tag(config=True)

    bib_command = List(["bibtex", "{filename}"], help="Shell command used to run bibtex.").tag(
        config=True
    )

    verbose = Bool(False, help="Whether to display the output of latex commands.").tag(config=True)

    texinputs = Unicode(help="texinputs dir. A notebook's directory is added")
    writer = Instance("nbconvert.writers.FilesWriter", args=(), kw={"build_directory": "."})

    output_mimetype = "application/pdf"

    _captured_output: List[str] = List()

    @default("file_extension")
    def _file_extension_default(self):
        return ".pdf"

    @default("template_extension")
    def _template_extension_default(self):
        return ".tex.j2"

    def run_command(  # noqa
        self, command_list, filename, count, log_function, raise_on_failure=None
    ):
        """Run command_list count times.

        Parameters
        ----------
        command_list : list
            A list of args to provide to Popen. Each element of this
            list will be interpolated with the filename to convert.
        filename : unicode
            The name of the file to convert.
        count : int
            How many times to run the command.
        raise_on_failure: Exception class (default None)
            If provided, will raise the given exception for if an instead of
            returning False on command failure.

        Returns
        -------
        success : bool
            A boolean indicating if the command was successful (True)
            or failed (False).
        """
        command = [c.format(filename=filename) for c in command_list]

        # This will throw a clearer error if the command is not found
        cmd = shutil.which(command_list[0])
        if cmd is None:
            link = "https://nbconvert.readthedocs.io/en/latest/install.html#installing-tex"
            msg = (
                f"{command_list[0]} not found on PATH, if you have not installed "
                f"{command_list[0]} you may need to do so. Find further instructions "
                f"at {link}."
            )
            raise OSError(msg)

        times = "time" if count == 1 else "times"
        self.log.info("Running %s %i %s: %s", command_list[0], count, times, command)

        shell = sys.platform == "win32"
        if shell:
            command = subprocess.list2cmdline(command)  # type:ignore[assignment]
        env = os.environ.copy()
        prepend_to_env_search_path("TEXINPUTS", self.texinputs, env)
        prepend_to_env_search_path("BIBINPUTS", self.texinputs, env)
        prepend_to_env_search_path("BSTINPUTS", self.texinputs, env)

        with open(os.devnull, "rb") as null:
            stdout = subprocess.PIPE if not self.verbose else None
            for _ in range(count):
                p = subprocess.Popen(
                    command,
                    stdout=stdout,
                    stderr=subprocess.STDOUT,
                    stdin=null,
                    shell=shell,  # noqa
                    env=env,
                )
                out, _ = p.communicate()
                if p.returncode:
                    if self.verbose:  # noqa
                        # verbose means I didn't capture stdout with PIPE,
                        # so it's already been displayed and `out` is None.
                        out_str = ""
                    else:
                        out_str = out.decode("utf-8", "replace")
                    log_function(command, out)
                    self._captured_output.append(out_str)
                    if raise_on_failure:
                        msg = f'Failed to run "{command}" command:\n{out_str}'
                        raise raise_on_failure(msg)
                    return False  # failure
        return True  # success

    def run_latex(self, filename, raise_on_failure=LatexFailed):
        """Run xelatex self.latex_count times."""

        def log_error(command, out):
            self.log.critical("%s failed: %s\n%s", command[0], command, out)

        return self.run_command(
            self.latex_command, filename, self.latex_count, log_error, raise_on_failure
        )

    def run_bib(self, filename, raise_on_failure=False):
        """Run bibtex one time."""
        filename = os.path.splitext(filename)[0]

        def log_error(command, out):
            self.log.warning(
                "%s had problems, most likely because there were no citations", command[0]
            )
            self.log.debug("%s output: %s\n%s", command[0], command, out)

        return self.run_command(self.bib_command, filename, 1, log_error, raise_on_failure)

    def from_notebook_node(self, nb, resources=None, **kw):
        """Convert from notebook node."""
        latex, resources = super().from_notebook_node(nb, resources=resources, **kw)
        # set texinputs directory, so that local files will be found
        if resources and resources.get("metadata", {}).get("path"):
            self.texinputs = os.path.abspath(resources["metadata"]["path"])
        else:
            self.texinputs = os.getcwd()

        self._captured_outputs = []
        with TemporaryDirectory() as td, _contextlib_chdir.chdir(td):
            notebook_name = "notebook"
            resources["output_extension"] = ".tex"
            tex_file = self.writer.write(latex, resources, notebook_name=notebook_name)
            self.log.info("Building PDF")
            self.run_latex(tex_file)
            if self.run_bib(tex_file):
                self.run_latex(tex_file)

            pdf_file = notebook_name + ".pdf"
            if not os.path.isfile(pdf_file):
                raise LatexFailed("\n".join(self._captured_output))
            self.log.info("PDF successfully created")
            with open(pdf_file, "rb") as f:
                pdf_data = f.read()

        # convert output extension to pdf
        # the writer above required it to be tex
        resources["output_extension"] = ".pdf"
        # clear figure outputs and attachments, extracted by latex export,
        # so we don't claim to be a multi-file export.
        resources.pop("outputs", None)
        resources.pop("attachments", None)

        return pdf_data, resources



from jupyter_client.manager import KernelManager


class FakeCustomKernelManager(KernelManager):
    expected_methods = {
        "__init__": 0,
        "client": 0,
        "start_kernel": 0,
    }

    def __init__(self, *args, **kwargs):
        self.log.info("FakeCustomKernelManager initialized")
        self.expected_methods["__init__"] += 1
        super().__init__(*args, **kwargs)

    def start_kernel(self, *args, **kwargs):
        self.log.info("FakeCustomKernelManager started a kernel")
        self.expected_methods["start_kernel"] += 1
        return super().start_kernel(*args, **kwargs)

    def client(self, *args, **kwargs):
        self.log.info("FakeCustomKernelManager created a client")
        self.expected_methods["client"] += 1
        return super().client(*args, **kwargs)



"""Tests for the HTMLSanitize preprocessor"""

from nbformat import v4 as nbformat

from nbconvert.preprocessors.sanitize import SanitizeHTML

from .base import PreprocessorTestsBase


class TestSanitizer(PreprocessorTestsBase):
    """Contains test functions for sanitize.py"""

    maxDiff = None

    def build_preprocessor(self):
        """Make an instance of a preprocessor"""
        preprocessor = SanitizeHTML()
        preprocessor.enabled = True
        return preprocessor

    def preprocess_source(self, cell_type, source, preprocessor):
        nb = self.build_notebook()
        res = self.build_resources()

        nb.cells[0].cell_type = cell_type
        nb.cells[0].source = source

        nb, res = preprocessor(nb, res)

        return nb.cells[0].source

    def test_constructor(self):
        """Can a SanitizeHTML be constructed?"""
        self.build_preprocessor()

    def test_svg_handling(self):
        """
        Test to make sure that svgs are handled 'properly'

        We only allow <img> tags (via markdown syntax) and not all the other ways
        to embed svg: <object>, <embed>, <iframe> nor inline <svg>
        """
        preprocessor = self.build_preprocessor()
        preprocessor.strip = True

        self.assertEqual(
            self.preprocess_source(
                "markdown",
                """
                ![some image](http://example.com/something.svg)

                <object data="something.svg" type="image/svg+xml"></object>

                <embed data="something.svg" type="image/svg+xml" />

                <iframe src="http://example.com/something.svg"></iframe>

                <svg xmlns="http://www.w3.org/2000/svg" viewBox="0 0 68 65">
                    <path fill="#1A374D" d="M42 27v-20c0-3.7-3.3-7-7-7s-7 3.3-7 7v21l12 15-7 15.7c14.5 13.9 35 2.8 35-13.7 0-13.3-13.4-21.8-26-18zm6 25c-3.9 0-7-3.1-7-7s3.1-7 7-7 7 3.1 7 7-3.1 7-7 7z"/>
                    <path d="M14 27v-20c0-3.7-3.3-7-7-7s-7 3.3-7 7v41c0 8.2 9.2 17 20 17s20-9.2 20-20c0-13.3-13.4-21.8-26-18zm6 25c-3.9 0-7-3.1-7-7s3.1-7 7-7 7 3.1 7 7-3.1 7-7 7z"/>
                </svg>
                """,
                preprocessor,
            ).strip(),
            """
            ![some image](http://example.com/something.svg)
            """.strip(),
        )

    def test_tag_allowlist_stripping(self):
        """Test tag allowlisting + stripping out offending tags"""
        preprocessor = self.build_preprocessor()
        preprocessor.strip = True

        self.assertEqual(
            self.preprocess_source(
                "markdown", "_A_ <em>few</em> <script>tags</script>", preprocessor
            ),
            "_A_ <em>few</em> tags",
        )

    def test_comment_stripping(self):
        """Test HTML comment stripping"""
        preprocessor = self.build_preprocessor()

        self.assertEqual(
            self.preprocess_source("markdown", "_A_ <em>few</em> <!-- tags -->", preprocessor),
            "_A_ <em>few</em> ",
        )

        preprocessor.strip_comments = False
        self.assertEqual(
            self.preprocess_source("markdown", "_A_ <em>few</em> <!-- tags -->", preprocessor),
            "_A_ <em>few</em> <!-- tags -->",
        )

    def test_attributes_allowlist(self):
        """Test style"""
        preprocessor = self.build_preprocessor()

        preprocessor.attributes["a"] = ["href", "title"]

        self.assertEqual(
            self.preprocess_source(
                "markdown", '<a href="link" rel="nofollow">Hi</a>', preprocessor
            ),
            '<a href="link">Hi</a>',
        )

    def test_style_allowlist(self):
        """Test style"""
        preprocessor = self.build_preprocessor()

        if "*" in preprocessor.attributes:
            preprocessor.attributes["*"].append("style")
        else:
            preprocessor.attributes["*"] = ["style"]
        preprocessor.styles = [
            "color",
        ]

        self.assertEqual(
            self.preprocess_source(
                "markdown",
                '_A_ <em style="color: blue; background-color: pink">'
                "few</em> <script>tags</script>",
                preprocessor,
            ),
            '_A_ <em style="color: blue;">few</em> &lt;script&gt;tags&lt;/script&gt;',
        )

    def test_tag_passthrough(self):
        """Test passing through raw output"""
        preprocessor = self.build_preprocessor()

        self.assertEqual(
            self.preprocess_source("raw", "_A_ <em>few</em> <script>tags</script>", preprocessor),
            "_A_ <em>few</em> &lt;script&gt;tags&lt;/script&gt;",
        )

    def test_output_sanitizing(self):
        """Test that outputs are also sanitized properly"""
        preprocessor = self.build_preprocessor()
        nb = self.build_notebook()

        outputs = [
            nbformat.new_output(
                "display_data",
                data={
                    "text/plain": "b",
                    "text/html": "<script>more evil</script>",
                    "text/css": "<style> * {display:none}</style>",
                },
            ),
            nbformat.new_output("stream", name="stdout", text="wat"),
            nbformat.new_output("stream", name="stdout", text="<script>Evil tag</script>"),
        ]
        nb.cells[0].outputs = outputs

        res = self.build_resources()
        nb, res = preprocessor(nb, res)

        expected_output = [
            {
                "data": {"text/html": "&lt;script&gt;more evil&lt;/script&gt;", "text/plain": "b"},
                "metadata": {},
                "output_type": "display_data",
            },
            {"name": "stdout", "output_type": "stream", "text": "wat"},
            {"name": "stdout", "output_type": "stream", "text": "<script>Evil tag</script>"},
        ]
        self.assertEqual(nb.cells[0].outputs, expected_output)

    def test_tag_allowlist(self):
        """Test tag allowlisting"""
        preprocessor = self.build_preprocessor()

        self.assertEqual(
            self.preprocess_source(
                "markdown", "_A_ <em>few</em> <script>tags</script>", preprocessor
            ),
            "_A_ <em>few</em> &lt;script&gt;tags&lt;/script&gt;",
        )



"""Tests for ScriptExporter"""

# Copyright (c) IPython Development Team.
# Distributed under the terms of the Modified BSD License.

import os
import sys

from nbformat import v4

import tests
from nbconvert.exporters.script import ScriptExporter

from .base import ExportersTestsBase


class TestScriptExporter(ExportersTestsBase):
    """Tests for ScriptExporter"""

    exporter_class = ScriptExporter  # type:ignore

    def test_constructor(self):
        """Construct ScriptExporter"""
        self.exporter_class()  # type:ignore

    def test_export(self):
        """ScriptExporter can export something"""
        (output, resources) = self.exporter_class().from_filename(  # type:ignore
            self._get_notebook()
        )
        assert len(output) > 0

    def test_export_python(self):
        """delegate to custom exporter from language_info"""
        self.exporter_class()  # type:ignore

        pynb = v4.new_notebook()
        (output, resources) = self.exporter_class().from_notebook_node(pynb)  # type:ignore
        self.assertNotIn("# coding: utf-8", output)

        pynb.metadata.language_info = {
            "name": "python",
            "mimetype": "text/x-python",
            "nbconvert_exporter": "python",
        }
        (output, resources) = self.exporter_class().from_notebook_node(pynb)  # type:ignore
        self.assertIn("# coding: utf-8", output)

    def test_export_config_transfer(self):
        """delegate config to custom exporter from language_info"""
        nb = v4.new_notebook()
        nb.metadata.language_info = {
            "name": "python",
            "mimetype": "text/x-python",
            "nbconvert_exporter": "python",
        }

        exporter = self.exporter_class()  # type:ignore
        exporter.from_notebook_node(nb)
        assert exporter._exporters["python"] != exporter
        assert exporter._exporters["python"].config == exporter.config


def test_script_exporter_entrypoint():
    nb = v4.new_notebook()
    nb.metadata.language_info = {
        "name": "dummy",
        "mimetype": "text/x-dummy",
    }

    p = os.path.join(os.path.dirname(tests.__file__), "exporter_entrypoint")
    sys.path.insert(0, p)
    try:
        output, _ = ScriptExporter().from_notebook_node(nb)
        assert output == "dummy-script-exported"
    finally:
        sys.path.remove(p)



"""
Contains Stdout writer
"""

# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

from nbconvert.utils import io

from .base import WriterBase


class StdoutWriter(WriterBase):
    """Consumes output from nbconvert export...() methods and writes to the
    stdout stream."""

    def write(self, output, resources, **kw):
        """
        Consume and write Jinja output.

        See base for more...
        """
        io.unicode_std_stream().write(output)



from nbconvert.filters import get_metadata


def test_get_metadata():
    output = {
        "metadata": {
            "width": 1,
            "height": 2,
            "image/png": {
                "unconfined": True,
                "height": 3,
            },
        }
    }
    assert get_metadata(output, "nowhere") is None
    assert get_metadata(output, "height") == 2
    assert get_metadata(output, "unconfined") is None
    assert get_metadata(output, "unconfined", "image/png") is True
    assert get_metadata(output, "width", "image/png") == 1
    assert get_metadata(output, "height", "image/png") == 3
